/**
 * @file
 * @brief  add 20211110
 */
#pragma once

#include <vector>
#include <list>
#include <array>
#include <set>
#include <string.h>
#include <cmath>
#include <cstdint>
#include <memory>
#include "PD_Input.h"
#include "PD_EgoVehStsArb.h"
#include "PD_LatTrajectory.h"
#include "../../source/PD_commonfunc.h"


#define FLAG_PREDICTION_TNT_BA true     /********* switch ***********/


namespace ADC
{
namespace PD_SVEstmd
{
// const float maxLateralSpeed = 1.5f;  // mps
// const float minLateralSpeed = -1.5f;  // mps
// const float maxValidtimeCoeff[3] = {20.763f, -21.667f, 7.0833f};    // {1.5, 1.1, 0.9}mps-{4.2, 5.5, 7}s
const float maxLateralSpeed = 1.0f;  // mps
const float minLateralSpeed = -1.0f;  // mps

/**
 * @class 
 * @brief 
 */
struct PD_SVEstmd_SV_msg
{
    uint32_t obj_ID=0;
    float s_SV = 0.0f;
    float ds_SV = 0.0f;
    float dds_SV = 0.0f;
    float dds_validtime_SV = 0.0f;
    float l_SV = 0.0f;
    float dl_SV = 0.0f;
    float dl_validtime_SV = 0.0f;
    float ddl_SV = 0.0f;
    float ddl_validtime_SV = 0.0f;

};
struct LS_point
{
    float t;    //unit-s
    float s_absolute;    //unit-m
    float s_relative;    //unit-m
    float l;    //unit-m
    float vs;    //unit-m/s
    float vl;    //unit-m/s
    // uint8_t lane_ID;
    LS_point()
    {
        t   =   0.0f;
        s_absolute   =   0.0f;
        s_relative   =   0.0f;
        l   =   0.0f;
        vs  =   0.0f;
        vl  =   0.0f;
    }
};
struct PD_SV_ValidTime_Next                 // valid time unit-s
{
    float dds_SV = 0.0f;
    float dds_validtime_SV = 0.0f;
    float dl_SV = 0.0f;
    float dl_validtime_SV = 0.0f;
    float ddl_SV = 0.0f;
    float ddl_validtime_SV = 0.0f;
};
struct PD_obj_SV                            // for one SV object
{
    uint32_t m_obj_ID=0;
    std::vector<LS_point> pd_ls_point_list;
    PD_SV_ValidTime_Next pd_SV_ValidTime_Next;
};

class SVEstmd
{
public:
    SVEstmd(const PD::PD_Input &m_pd_input);
    ~SVEstmd()=default;

private:
    bool init(const PD::PD_Input &m_pd_input);
    bool initFlag=true;
    // void getSVData();
    std::vector<PD_SVEstmd_SV_msg> _pd_SVEstmd_SV_msg_list;
    std::vector<ADC::PD_EgoVehTrj::EgoVeh_TrjPnt> _pd_SVEstmd_EV_LgtTraj;          //ego vehicle lgt
    bool _is_lgtTraj_directCallIndex = false;                                       //ego vehicle lgt
    float _BA_duration_time_input=0.0;
    uint32_t _passage_corridor_ID=0;
    LineGeometry _BA_corridor_line;
    float linearInterpolation(const float x, const std::pair<float,float>& point1, const std::pair<float,float>& point2);

public:
    void getLgtTrajectoryPoint(const std::vector<ADC::PD_EgoVehTrj::EgoVeh_TrjPnt> &m_get_LgtTraj);
    void process();                                                 // working
    std::vector<PD_obj_SV> SV_Trajectory_list;                      // output

};






}       //namespace PD_SVEstmd
}       //namespace ADC











#if FLAG_PREDICTION_TNT_BA
#define OVERFLOWED 1E-12
namespace pd_forwardSpread
{

/**********************************************
restriction
**********************************************/
constexpr float x_max_objectNeedPredict = 40.0f;
constexpr float y_min_objectNeedPredict = 10.0f;
constexpr float vx_highway_objectNeedPredict = 60.0f/3.6f;  // only if speed is larger than 60kph
constexpr float vx_urban_objectNeedPredict = 15.0f/3.6f;  // only if speed is larger than 15kph
constexpr int n_trajPoints_Cartesian = 20;// necessary tnt trigger input
constexpr int n_trajPoints_Frenet = 5;// 20220729 use five points 0 5 10 15 19
constexpr int n_intentionNumber = 1;// 20220818 the number of intention default 1
constexpr float thresholdValue_intention = 0.02f;// 20220823 set&reduce threshold of intention if use multi-intentions

/**********************************************
classification
input 1*10
Wa 10*12    Ba 1*12
Wb 12*9     Bb 1*9
Wc 9*120    Bc 1*120
**********************************************/
constexpr int n_row_input_class = 1;                constexpr int n_col_input_class = 10;
constexpr int n_row_Wa_class = n_col_input_class;   constexpr int n_col_Wa_class = 12;
constexpr int n_row_Ba_class = n_row_input_class;   constexpr int n_col_Ba_class = n_col_Wa_class;
constexpr int n_row_Wb_class = n_col_Wa_class;      constexpr int n_col_Wb_class = 9;
constexpr int n_row_Bb_class = n_row_input_class;   constexpr int n_col_Bb_class = n_col_Wb_class;
constexpr int n_row_Wc_class = n_col_Wb_class;      constexpr int n_col_Wc_class = 120;
constexpr int n_row_Bc_class = n_row_input_class;   constexpr int n_col_Bc_class = n_col_Wc_class;

/**********************************************
regression
input 1*12
Wa 12*14    Ba 1*14
Wb 14*10    Bb 1*10
Wc 10*12    Bc 1*12
**********************************************/
constexpr int n_row_input_regr = 1;               constexpr int n_col_input_regr = 12;
constexpr int n_row_Wa_regr = n_col_input_regr;   constexpr int n_col_Wa_regr = 14;
constexpr int n_row_Ba_regr = n_row_input_regr;   constexpr int n_col_Ba_regr = n_col_Wa_regr;
constexpr int n_row_Wb_regr = n_col_Wa_regr;      constexpr int n_col_Wb_regr = 10;
constexpr int n_row_Bb_regr = n_row_input_regr;   constexpr int n_col_Bb_regr = n_col_Wb_regr;
constexpr int n_row_Wc_regr = n_col_Wb_regr;      constexpr int n_col_Wc_regr = 12;
constexpr int n_row_Bc_regr = n_row_input_regr;   constexpr int n_col_Bc_regr = n_col_Wc_regr;


/**********************************************
lane
**********************************************/
constexpr float y_laneCenter[7] = {10.5, 7.0, 3.5, 0, -3.5, -7, -10.5};


class Matrix
{
public:
    Matrix(int m, int n);
    Matrix(int n);
    Matrix(const Matrix &);
    Matrix(float* items, int m, int n);
    ~Matrix();
    static Matrix FromFile(std::string file);
    int getRowNum() const;
    int getColNum() const;

    Matrix Trans() const;

    float get(int i, int j) const;
    void set(int i, int j, float val);

    Matrix operator +(const Matrix &m);
    Matrix operator -(const Matrix &m);
    Matrix operator *(const Matrix &m);
    Matrix operator *(const float f);
    Matrix& operator=(const Matrix& m);
    Matrix Inverse();
    
    // friend std::ostream& operator <<(std::ostream &os, const Matrix &m);
    Matrix operator <<(const std::vector<std::vector<float> > &m); //vector赋值
    void vector2matrix(const std::vector<std::vector<float> > &m);
    std::vector<std::vector<float> > matrix2vector();
    void showMat() const;

private:
    float *item;
    int rowNum;
    int colNum;

private:
    void RowSwap(int i, int j, float multiply);
    void RowSwap(int i, int j);
    void FlowOver();
};


class ObjectPredictionTNT
{
public:
    enum mode_type{
        HIGHWAY,
        URBAN,
        MOTORCYCLE,
        TRICYCLE,
        PEDESTRIAN
    };
public:
    ObjectPredictionTNT(const mode_type mode);
    ~ObjectPredictionTNT()=default;
public:
    void getHistoryTraj(std::vector<std::vector<float> > historyTraj);
    void prediction();

private:
    Matrix relu(const Matrix &m);               // 0
    Matrix softmax(const Matrix &m);            // e%
    std::vector<std::vector<float> > label_decode_firstlayer(const std::vector<std::vector<float> > &m);

private:
    mode_type _mode;
    bool isValid = true;
    void coeffInit();
    void classification();
    void regression();
    void intentionPrediction();
    void trajectoryprediction();
    void postProcess();

public:
    static constexpr float delta_t_input = 0.1f, delta_t_output = 0.5f;
private:
    float _s_cur = 0.0f, _l_cur = 0.0f;
    // std::vector<std::vector<float> > input(input_row,std::vector<float> (input_col));
    // std::vector<std::vector<float> > Wa(Wa_row,std::vector<float> (Wa_col));
    // std::vector<std::vector<float> > Ba(Ba_row,std::vector<float> (Ba_col));
    // std::vector<std::vector<float> > Wb(Wb_row,std::vector<float> (Wb_col));
    // std::vector<std::vector<float> > Bb(Bb_row,std::vector<float> (Bb_col));
    // std::vector<std::vector<float> > Wc(Wc_row,std::vector<float> (Wc_col));
    // std::vector<std::vector<float> > Bc(Bc_row,std::vector<float> (Bc_col));
    // std::vector<std::vector<float> > input;
    // std::vector<std::vector<float> > Wa;
    // std::vector<std::vector<float> > Ba;
    // std::vector<std::vector<float> > Wb;
    // std::vector<std::vector<float> > Bb;
    // std::vector<std::vector<float> > Wc;
    // std::vector<std::vector<float> > Bc;
    std::unique_ptr< std::vector<std::vector<float> > > _classInput;
    std::shared_ptr< std::vector<std::vector<float> > > _classWa;
    std::shared_ptr< std::vector<std::vector<float> > > _classBa;
    std::shared_ptr< std::vector<std::vector<float> > > _classWb;
    std::shared_ptr< std::vector<std::vector<float> > > _classBb;
    std::shared_ptr< std::vector<std::vector<float> > > _classWc;
    std::shared_ptr< std::vector<std::vector<float> > > _classBc;
// public:
    std::shared_ptr< std::vector<std::vector<float> > > _classOutput;

private:
    std::unique_ptr< std::vector<std::vector<float> > > _regrInput;
    std::shared_ptr< std::vector<std::vector<float> > > _regrWa;
    std::shared_ptr< std::vector<std::vector<float> > > _regrBa;
    std::shared_ptr< std::vector<std::vector<float> > > _regrWb;
    std::shared_ptr< std::vector<std::vector<float> > > _regrBb;
    std::shared_ptr< std::vector<std::vector<float> > > _regrWc;
    std::shared_ptr< std::vector<std::vector<float> > > _regrBc;
// public:
    std::unique_ptr< std::vector<std::vector<float> > > _regrOutput;    // 1*12

    std::shared_ptr< std::vector<std::vector<float> > > objectOutput;   // n*12
public:
    std::shared_ptr< std::vector<std::vector<float> > > getIntentionOutput() const {return _classOutput;}
    std::shared_ptr< std::vector<std::vector<float> > > getObjectOutput() const {return objectOutput;}
};








struct S_obstacle_tral{
    float t_s;
    float x_m;
    float y_m;
    float vx_m;
    float vy_m;
};

class ObstaclePrediction{   // for all obstacles
public:
    ObstaclePrediction(){}
    ~ObstaclePrediction(){}

public:
    // bool isValid = true;
    void process();


// histroy trajectory
public:
    void saveSurround(const RoadInfo &m_rm_msg);
    void saveEgoVehicle(const dpc::VED4DPC &m_vse_msg);
private:    // helper function for saving histroy trajectory
    float _rate_angle = 0.0f;  // rad/s
    float _velo_x = 0.0f;      // m/s
    float _velo_y = 0.0f;      // m/s
    void saveObstacleHist(const Object &m_object, const float m_delta_t_rm, const float m_d_angle, const float m_d_x, const float m_d_y);
    std::unique_ptr<
        std::map< uint32_t, std::list<S_obstacle_tral> > >   
        _obstacles_histCartesianTraj_timelist = std::make_unique<      // 0.1s interval
        std::map< uint32_t, std::list<S_obstacle_tral> > >();
    void updateObstaclePos(const Object &m_object, const float m_d_angle, const float m_d_x, const float m_d_y);
    void recordExistedObstaclePos(const Object &m_object, const float m_delta_t_rm);
    void recordNonexistedObstaclePos(const Object &m_object);
    void recordLastObstaclePos(const Object &m_object, const float m_delta_t_rm);
    std::unique_ptr< 
        std::map< uint32_t, S_obstacle_tral> >
        _obstacles_lastCartesianPos = std::make_unique<                 // save the last position, no fixed interval
        std::map< uint32_t, S_obstacle_tral> >();
    void eraseFirstObstaclePos(const Object &m_object);
    void clearLossedObstacle();
    std::unique_ptr<
        std::map< uint32_t, bool> >
        _obstacles_updated_inThisCycle = std::make_unique<
        std::map< uint32_t, bool> >();
    float linearInterpolation(const float x, const std::pair<float,float>& point1, const std::pair<float,float>& point2);
    float linearLookupTable(const float x, const std::list<float>& table_x, const std::list<float>& table_y);


// coordinate transformation
private:
    void saveCorridor(const RoadInfo &m_rm_msg);
    Corridor linearExtendCorridor(const Corridor &m_corridor, const float m_startRange, float m_endRange, float m_leftGradient=std::nanf("1"));
    std::unique_ptr<
        std::list<std::list<Corridor> > >
        _corridors_inLanes = std::make_unique<
        std::list<std::list<Corridor> > >();
    void calculateFrenetTraj();
    // std::unique_ptr<
    //     std::list<std::tuple< uint32_t, std::list<Corridor>, std::vector<std::vector<float>> > > >/* object_id, corridors, traj*/
    //     _obstacles_histFrenetTraj_forward = std::make_unique<                                     // for forward
    //     std::list<std::tuple< uint32_t, std::list<Corridor>, std::vector<std::vector<float>> > > >();
    std::unique_ptr<
        std::map< uint32_t, std::pair<std::list<Corridor>, std::vector<std::vector<float>> > > >/* object_id, corridors, traj20*2 */
        _obstacles_histFrenetTraj_forward = std::make_unique<                                     // for forward
        std::map< uint32_t, std::pair<std::list<Corridor>, std::vector<std::vector<float>> > > >();
    std::pair<float, float> _c0_leftAndRight;
// public:
//     std::unique_ptr< std::map< uint32_t, std::pair<std::list<Corridor>, std::vector<std::vector<float>> > > > get_obstacles_histFrenet() const
//         {return _obstacles_histFrenetTraj_forward;}

private:    // helper function for coordinate transformation
    std::list<Corridor> findNearestCorridors(const S_obstacle_tral& m_obstacle_point);
    std::vector<float> CartesianToFrenet(const std::list<Corridor>& m_corridor_list, const float m_x, const float m_y);
    float calculateYPolynomial(const Corridor& m_corridor, const float x);
    float calculateYPolynomial(const std::list<Corridor>& m_corridors, const float x);
    float calculateDisBetweenPoints(const std::pair<float,float>& point1, const std::pair<float,float>& point2);
    void findReferPointNewtonMethod(float& refer_x0, float& refer_x1, const float m_x, const float m_y, 
                                    const std::list<Corridor>& m_corridors,
                                    const float step_x=0.1f, const float accuracy_diff=0.02f, const int cycleNum=60);
    float calculateArcLengthAbs(const std::list<Corridor>& m_corridors, const float refer_x);
    float calculateArcIntegralSimpsonMethod(const Corridor& m_corridor, const float refer_x);
    float calculateDeltaArc(const Corridor& m_corridor, const float point_x);
    void fitterFrenetHistTraj(std::vector<std::vector<float> > &m_trajVector);


// prediction
private:
    std::unique_ptr<ObjectPredictionTNT> _tntHighway = std::make_unique<ObjectPredictionTNT>(pd_forwardSpread::ObjectPredictionTNT::mode_type::HIGHWAY);
    std::unique_ptr<ObjectPredictionTNT> _tntUrban = std::make_unique<ObjectPredictionTNT>(pd_forwardSpread::ObjectPredictionTNT::mode_type::URBAN);
    // std::unique_ptr<ObjectPredictionTNT> _tntMotorcycle = std::make_unique<ObjectPredictionTNT>(pd_forwardSpread::ObjectPredictionTNT::mode_type::MOTORCYCLE);
    // std::unique_ptr<ObjectPredictionTNT> _ntTricycle = std::make_unique<ObjectPredictionTNT>(pd_forwardSpread::ObjectPredictionTNT::mode_type::TRICYCLE);
    // std::unique_ptr<ObjectPredictionTNT> _tntPedestrian = std::make_unique<ObjectPredictionTNT>(pd_forwardSpread::ObjectPredictionTNT::mode_type::PEDESTRIAN);
private:
    void restrictPredictL();
    std::unique_ptr<
        std::map< uint32_t, std::list<int> > >  /* object_id, history_intention 0 1 2 */
        _obstacles_predictionHistoryIntention = std::make_unique<
        std::map< uint32_t, std::list<int> > >();
private:
    void predictTrajectory();
    std::unique_ptr<
        std::map< uint32_t, std::vector<std::vector<float>> > >/* object_id, traj n*12 */
        _obstacles_predictionTraj = std::make_unique<
        std::map< uint32_t, std::vector<std::vector<float>> > >();
    void translateLToEgo();
    std::shared_ptr<
        std::map< uint32_t, std::vector<std::vector<float>> > >/* object_id, traj n*12 */
        _obstacles_predictionTraj_toEgo = std::make_shared<     // l is translated to ego vehicle, not to corridor
        std::map< uint32_t, std::vector<std::vector<float>> > >();

public:
    std::shared_ptr< std::map< uint32_t, std::vector<std::vector<float>> > > get_obstacles_predictionTraj_toEgo() const 
        {return _obstacles_predictionTraj_toEgo;}


// data
private:
    void savefile();


};



#define has_member_inStruct(findMember) \
template<typename T>\
struct has_member_inStruct_##findMember{\
    template <typename _T>static auto check(_T)->typename std::decay<decltype(_T::findMember)>::type;\
    static void check(...);\
    using type=decltype(check(std::declval<T>()));\
    enum{value=!std::is_void<type>::value};\
};
has_member_inStruct(vse_velocity_ips);


}   // namespace pd_forwardSpread

#endif  // FLAG_PREDICTION_TNT_BA

#if FLAG_PREDICTION_TNT_BA
extern std::unique_ptr<pd_forwardSpread::ObstaclePrediction> obstaclePrediction;
#endif